/* regarder la fonction get_user_pages() dans mm/memory.c */
/* regarder la fonction access_process_vm dans kernel/ptrace.c */
 
#define __KERNEL__
#define MODULE

#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/sys.h>
#include <linux/smp_lock.h>
#include <linux/unistd.h>
#include <linux/highmem.h>

struct mem_addr {
        unsigned long vm_start;
        unsigned long vm_end;
        unsigned long vm_start_cache;
        unsigned long vm_end_cache;
        int map_count;
        unsigned long startcode;
        unsigned long endcode;
        unsigned long startdata;
        unsigned long enddata;
        unsigned long startbrk;
        unsigned long startstack;
        unsigned long arg_start;
        unsigned long arg_end;
        unsigned long env_start;
        unsigned long env_end;
        unsigned long rss;
        unsigned long total_vm;
        unsigned long locked_vm;
        unsigned long swap_address;
};

int (*handle_mm_fault_e)( struct mm_struct *, struct vm_area_struct *, 
	unsigned long , int );
struct vm_area_struct *(*find_extend_vma_e)( struct mm_struct *, 
	unsigned long );
extern void* sys_call_table[];

struct 
task_struct * get_task( pid_t pid )
{
	struct task_struct * p;
	struct task_struct ** htable = &pidhash[ pid_hashfn(pid) ];

	read_lock( &tasklist_lock );
	for ( p = *htable; p && p->pid != pid; p = p->pidhash_next );
	read_unlock( &tasklist_lock );
	
	return( p );
}

int 
getvirtaddr( struct mem_addr * temmem, int pid )
{ 
	struct task_struct * temp;
	temp = get_task( pid );
	if ( temp == NULL ) {
		return( 1 );
  	}
	if ( temp->mm == NULL ) {
		return( 1 );
	}
	if ( temp->mm->mmap != NULL ) {	
		temmem->vm_start = temp->mm->mmap->vm_start;
		temmem->vm_end = temp->mm->mmap->vm_end;
	} else {
		temmem->vm_start = 0;
		temmem->vm_end = 0;
	}
	if ( temp->mm->mmap_cache != NULL ) {
		temmem->vm_start_cache = temp->mm->mmap_cache->vm_start;
		temmem->vm_end_cache = temp->mm->mmap_cache->vm_end;
	} else {
		temmem->vm_start_cache = 0;
		temmem->vm_end_cache =0;
	}
	temmem->map_count = temp->mm->map_count;
	temmem->startcode = temp->mm->start_code;
	temmem->endcode = temp->mm->end_code;
	temmem->startdata = temp->mm->start_data;
	temmem->enddata = temp->mm->end_data;
	temmem->startbrk = temp->mm->start_brk;
	temmem->startstack = temp->mm->start_stack;
	temmem->arg_start = temp->mm->arg_start;
	temmem->arg_end = temp->mm->arg_end;
	temmem->env_start = temp->mm->env_start;
	temmem->env_end = temp->mm->env_end;
	temmem->rss = temp->mm->rss;
	temmem->total_vm = temp->mm->total_vm;
	temmem->locked_vm = temp->mm->locked_vm;
#ifndef _RH_
	temmem->swap_address = temp->mm->swap_address;
#endif
	return( 0 );
}

struct page * 
getpagefromaddr( unsigned long addr, struct task_struct * task ) 
{ 
	pgd_t * pgd;
	pmd_t * pmd;
	pte_t * pte;
	struct vm_area_struct * vma;
	struct mm_struct * mm;
 
startpoint:
	task_lock( task );
	mm = task->mm;
	if ( mm ) atomic_inc( &mm->mm_users );
	task_unlock( task );
	if ( !mm ) return( NULL );

	spin_lock( &task->mm->page_table_lock );
  
	/* pgd */
	pgd = pgd_offset( mm, addr );
	if ( pgd_none(*pgd) ) {
		goto fault;
	}
	if ( pgd_bad(*pgd) ) {
		pgd_ERROR( *pgd );
		goto error;
	}

	/* pmd */
	pmd = pmd_offset( pgd, addr );
	if ( pmd_none(*pmd) ) {
		goto fault;
	}
	if ( pmd_bad(*pmd) ) {
		pmd_ERROR( *pmd );
		goto error;
	} 
 
	/* pte */
#ifndef _RH_
	pte = pte_offset( pmd, addr );
#else
	pte = pte_offset_kernel( pmd, addr ); /* ?? */
#endif
	if ( !pte_present(*pte) ) {
		goto fault;
	}

	spin_unlock( &task->mm->page_table_lock );
	return( pte_page(*pte) );

fault:
	spin_unlock( &task->mm->page_table_lock );
	vma = find_extend_vma_e( mm, addr );
	atomic_dec( &mm->mm_users );
	if ( handle_mm_fault_e(mm, vma, addr, 0) > 0 ) { 
		goto startpoint;
	} else {
		return( NULL );
	}
error:
	spin_unlock( &task->mm->page_table_lock );
	atomic_dec( &mm->mm_users );
	return( NULL ); 
} 

int 
rwfromvirt( int pid, char * buffer, int len, unsigned long addr, int w )
{ 
	struct task_struct * task;
	struct page * page;
	struct mm_struct * mm;
	char * maddr;
	
	if ( addr == 0 ) {
		return( -1 );
	}

	task = get_task( pid );
	if ( task == NULL ) return( 1 );
	page = getpagefromaddr( addr, task );
	if ( page == NULL ) return( -1 );
 
	if ( w ) {
		if ( (!VALID_PAGE(page)) || PageReserved(page) ) return( -2 );
		maddr = kmap( page );
		memcpy( maddr + (addr & ~PAGE_MASK), buffer, len );
		flush_page_to_ram( page );
		flush_icache_page( find_extend_vma_e(task->mm, addr), page );
		kunmap( page );
	} else {
		maddr = (char *)kmap(page);
		memcpy(buffer, maddr + (addr & ~PAGE_MASK), len);
		flush_page_to_ram(page);
		kunmap(page);
	}
	mm = task->mm;
	atomic_dec( &mm->mm_users );
	return( 0 );
}

#define __NR_getvirtaddr 251
#define __NR_rwfromvirt 250
static void * old_get = NULL;
static void * old_rw = NULL;

int 
init_module() 
{ 
	old_get = sys_call_table[ __NR_getvirtaddr ];
	sys_call_table[ __NR_getvirtaddr ] = getvirtaddr;
	old_rw = sys_call_table[ __NR_rwfromvirt ];
	sys_call_table[ __NR_rwfromvirt ] = rwfromvirt;
	return( 0 );
} 

void 
cleanup_module()
{
	sys_call_table[ __NR_getvirtaddr ] = old_get;
	sys_call_table[ __NR_rwfromvirt ] = old_rw;
} 

MODULE_LICENSE( "GPL" );
