/* 
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/mman.h>
#include <elf.h>
#include "../libdisasm/libdis.h"
#include "include/kernsh.h"

/* Ripped from my own works so, under GPL V2 :) - Kstat */
/* Some macros */
#define DTYPE() i.destType
#define CODE() i

int
get_jmp_off( struct instr i )
{
	if ( (DTYPE () & OP_TYPE_MASK) != (OP_REL) ) {
		return( 0 );
	}

	/* The operand is a byte  */
	if ( (DTYPE() & (OP_SIZE_MASK)) == OP_BYTE ) {
		if ( ((DTYPE() & (OP_MOD_MASK)) & (OP_SIGNED)) == (OP_SIGNED) ) {
			if ( strtol(CODE().dest, NULL, 16) + CODE().size > 127 ) {    /* 2^7 - 1 */
				return( strtol(CODE().dest, NULL, 16)+CODE().size-256 );
			} else {
				return( strtol(CODE().dest, NULL, 16)+CODE().size );
			}
		}
	}

	if ( (DTYPE() & (OP_SIZE_MASK)) == OP_HWORD ) {
		if ( ((DTYPE() & (OP_MOD_MASK)) & (OP_SIGNED)) == (OP_SIGNED) ) {
			if ( strtol(CODE().dest, NULL, 16) + CODE().size > 32767 ) {  /* 2^15 - 1 */
				return( strtol(CODE().dest, NULL, 16)+CODE().size-65536 );
			} else {
				return( strtol(CODE().dest, NULL, 16)+CODE().size );
			}
		}
	}

	if ( (DTYPE() & (OP_SIZE_MASK)) == OP_WORD ) {
		if ( ((DTYPE() & (OP_MOD_MASK)) & (OP_SIGNED)) == (OP_SIGNED) ) {
			if ( strtol(CODE().dest, NULL, 16) + CODE ().size > 2147483647 ) { /* 2^31 - 1 */
				return( strtol(CODE().dest, NULL, 0)+CODE().size-4294967296 );
			} else {
				return( strtol(CODE().dest, NULL, 16)+CODE().size );
			}
		}
	}

	/* Never see better :) */
	return( strtol(CODE().dest, NULL, 0)+CODE().size );
}

/* I have just modified a little bit the dissassembler to make it more
 * effecient with the addresses - Kstat
 */
void
display_code_asm( size_t start_address, unsigned char * buf, int len )
{
	unsigned char symbol_name[ 2*sizeof(size_t)+3 ];
	struct instr curr_inst;
	size_t p_symbol;
	int i = 0;
	int size;
	int n;
	int j = 0;
	int c;

	disassemble_init( IGNORE_NULLS, ATT_SYNTAX );

	while ( i < len ) {
		memset( &curr_inst, 0, sizeof(struct code) );
		fprintf( stdout, "%#x\t", start_address+i );
		size = disassemble_address( buf+i, &curr_inst );
		if ( size ) {
			for ( n = 0; n < 12; n++ ) {
				if ( n < size ) {
					fprintf( stdout, "%02x ", buf[i+n] );
				} else {
					fputs( "   ", stdout );
				}
			}
			fprintf( stdout, "%s", curr_inst.mnemonic );
			if ( curr_inst.dest[0] != 0 ) {
				/* Crazy isn't it ? Yes I learnt a lot making my own 
				 * disassembler ;) - Kstat 
				 */
				if ( (curr_inst.mnemType & (INS_TYPE_MASK)) == (INS_CALL) || (curr_inst.mnemType & (INS_TYPE_MASK)) == (INS_BRANCH) || (curr_inst.mnemType & (INS_TYPE_MASK)) == (INS_BRANCHCC) ) {
					if ( (curr_inst.destType & (OP_TYPE_MASK)) == (OP_REL) 
						&& (curr_inst.destType & (OP_PERM_MASK)) == (OP_X) ) {
						p_symbol = ( start_address + i ) + get_jmp_off( curr_inst );
						snprintf( symbol_name, sizeof(symbol_name), "%#x", p_symbol );
						fprintf( stdout, "\t%s (r)", get_symbol_by_addr(symbol_name) );
					} else {
						/* We have a direct call (a register ?) */
						if ( (curr_inst.destType & (OP_TYPE_MASK)) == (OP_REG) ) {
							fprintf( stdout, "\t%s (d reg)", curr_inst.dest );
						} else {
							fprintf( stdout, "\t%p (d)", 
								(void *)strtoul(curr_inst.dest,NULL,16) );
						}
					}
				} else {
					fprintf( stdout, "\t%s", curr_inst.dest );
				}
			}
			if ( curr_inst.src[0] != 0 ) fprintf( stdout, ",\t%s", curr_inst.src );
			if ( curr_inst.aux[0] != 0 ) fprintf( stdout, " %s", curr_inst.aux );
			fputs( "\n", stdout );
			i += size;
		} else {
			fatal( "invalid opcode %02x\n", buf[i] );
			i++;
		}

		if ( !script_on_command_line ) {
			j++;
			if ( !((j+1)%60) ) {
				c = enter();
				if ( c == '\r' || c == '\n' ) {
					continue;
				}
				if ( c == 'q' ) {
					getchar();
					return;
				}
			}
		}
	}
	return;
}

/* adapted from quikdis.c in libdisasm */
int
disasm( size_t start_address, int len )
{
	Elf32_Ehdr * target_header;
	Elf32_Phdr * program_header;
	struct stat tmpstat;
	void * image_vmlinux;
	unsigned char * buf = NULL;
	int entry;
	int ftarget;
	int n;
	int x;
	int offset;

	/* read ELF header */
	ftarget = found_vmlinux();
	if ( ftarget < 0 ) {
		return( -2 );
	}

	n = fstat( ftarget, &tmpstat );
	if ( n < 0 ) {
		perror( "disasm:fstat()" );
		return( -1 );
	}

	image_vmlinux = mmap( 0, tmpstat.st_size, PROT_READ, MAP_SHARED, 
	ftarget, 0 );
	if ( (int)image_vmlinux < 1 ) {
		perror( "disasm:mmap()" );
		return( -1 );
	}
	close( ftarget );

	target_header = image_vmlinux;

	/* read program header table */
	for ( x = 0; x < target_header->e_phnum; x++ ) {
		program_header = image_vmlinux + target_header->e_phoff + 
		( x * target_header->e_phentsize );

		/* If entry point is in this section */
		if ( target_header->e_entry >= program_header->p_vaddr && 
		target_header->e_entry <= 
		(program_header->p_vaddr + program_header->p_filesz)) {

			/* resolve entry point RVA to a file offset */
			entry = target_header->e_entry - 
			( program_header->p_vaddr - program_header->p_offset );

			/* read entire program segment into buffer */
			offset = start_address - target_header->e_entry;
			buf = image_vmlinux + entry + offset;
	
			display_code_asm( target_header->e_entry+offset, buf, len );
		}
	}

	munmap( image_vmlinux, tmpstat.st_size );
	disassemble_cleanup();
	return( 0 );
}
