#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "include/kernsh.h"

#define VAR_SIZE   16

static struct Env {
	char name[VAR_SIZE];
	struct Env * next;
	char * value;
} env;

static FILE * old_std = NULL;
static int pipe_what = 0;

char *
parse_line( char * line )
{
	char * c = line;
	char * f = NULL;
	char variable[VAR_SIZE];
	struct Env * first = NULL;
	int diff = 1;
	char * new_line = NULL;
	char * var = NULL;
	FILE * pipe = NULL;
	old_std = NULL;

	if ( line == NULL ) {
		return NULL;
	}

	new_line = dupstr( line );
	/* 
	char * line_tmp = dupstr ( line );
	c = line_tmp;
	*/

	while ( (c = strchr(c , '$')) != NULL ) {
		int size = 0;

		first = &env;
		if ( c > line && *(char *)(c-1) == '\\' ) {
			/* escaped string */
			c = c + 1;
			continue;
		}

		strncpy ( variable, c+1, sizeof(variable) );

		if ( (f = strchr(variable, ' ')) != NULL ) {
			*f = '\0';
		}

		var = dupstr( variable );
		if ( variable[0] == '{' && (f = strchr(variable,'}')) != NULL ) {
			*f = '\0';
			var[ f-variable+1 ]='\0';
			strncpy( variable, variable+1, sizeof(variable) );
		}

		new_line = realloc( new_line, (int)(c-line)+diff );
		if ( new_line == NULL ) {
			perror( "parse_line:realloc()" );
			return( NULL );
		}

		new_line[ (int)(c-line)+diff-1 ] = '\0';
		do {
			if ( *(first->name) && strcmp (first->name,variable ) == 0 ) {
				size = strlen( new_line );
				new_line = realloc( new_line, size+strlen(first->value)+1 );
				if ( new_line == NULL ) {
					perror( "parse_line:realloc()" );
					return( NULL );
				}
				strcpy( new_line+size, first->value );
				diff += strlen( new_line ) - size - strlen( var ) - 1;
				goto next_part;
			}
		} while ( (first=first->next) != NULL );

		size = strlen( new_line );
		new_line = realloc( new_line, strlen(c)+size );
		if ( new_line == NULL ) {
			perror( "parse_line:realloc()" );
			return( NULL );
		}

		strcpy( new_line+size, c++ );
		free( var );
		continue;

next_part:
		size = strlen( new_line );
		new_line = realloc( new_line, strlen(c)+size );
		if ( new_line == NULL ) {
			perror( "parse_line:realloc()" );
			return( NULL );
		}

		new_line[size] = ' ';
		strcpy( new_line+size , ++c+strlen(var) );
		free ( var );
	}

	/* fprintf( stdout, "Command : %s\n", new_line ); */
	c = line;
	line = new_line;
	free( c );

	if ( (c = strchr(new_line,'>')) != NULL || (c = strchr(new_line,'|')) != NULL ) {
		pipe_what = ( *c == '>' ) ? 1 : 2;
		if ( *(c-1) == '\\' ) {
			goto exit;
		}

		*c = '\0';
		c = stripwhite( ++c );
		if ( *c == '>' ) {
			*c++ = '\0';
			c = stripwhite( c );
			if ( (pipe = fopen(c,"a")) == NULL ) {
				perror( "parse_line:fopen()" );
				goto exit;
			}
		} else {
			if ( pipe_what == 1 ) {
				if ( (pipe = fopen(c,"w")) == NULL ) {
					perror( "parse_line:fopen()" );
					goto exit;
				}
			} else {
				if ( (pipe = popen(c,"w")) == NULL ) {
					perror( "parse_line:popen()" );
					goto exit;
				}
			}
		}

		old_std = stdout;
		stdout = pipe;

		/* then put the command_on_script system */
		script_on_command_line++;
	}

exit:
	return( (char *)new_line );
}

void
end_parse_line( void )
{
	if ( old_std != NULL ) {
		FILE * pipe = NULL;

		script_on_command_line--;
		pipe = stdout;
		stdout = old_std;

		if ( pipe_what == 1 ) {
			fclose( pipe );
		} else {
			pclose( pipe );
		}
		old_std = NULL;
	}
	return;
}

void
add_env( char * variable, char * value )
{
	struct Env * new = NULL;

	new = (struct Env *)malloc( sizeof(struct Env) );
	if ( new == NULL ) {
		perror( "add_env:malloc()" );
		return;
	}

	if ( variable == NULL || value == NULL ) {
		fatal( "add_env : There is a problem with yours values !\n" );
		return;
	}

	memset( new, 0, sizeof(struct Env) );
	if ( strlen(variable) != 0 ) {
		strncpy( new->name, variable, VAR_SIZE-1 );
		new->name[ VAR_SIZE-1 ] = '\0';
		new->next = env.next;
		new->value = value;
		env.next = new;
	} else {
		fatal( "Invalid Variable name !\n" );
	}
	return;
}

void
clear_all( struct Env * del )
{
	if ( del == NULL ) {
		return;
	}

	free( del->value );
	clear_all( del->next );
	free( del->next );
	return;
}

void
list_all( void )
{
	struct Env * first = env.next;
	
	if ( first == NULL ) {
		fputs( "No environment variables\n", stdout );
		return;
	}

	fputs( "[ Environment variables ]\n\n", stdout );
	while ( first != NULL ) {
		fprintf( stdout, "%s=%s\n", first->name, first->value );
		first = first->next;
	}
	return;
}

void
usage_env( void )
{
	fputs(
	"env [ -cls ] <arguments> \n\n"
	"This command permits to add some basic \"Environment\" variables\n"
	"allowing you to save some values and reuse them after.\n\n"
	"__options\n"
	"  -s [ name ] [value with space if you want ] : set a value of the\n"
	"environment. For example => set -s KMEM kmem -d 0xc0000000 96.\n"
	"  -l : list all environment variables ;\n"
	"  -c : clear *all* environment variables.\n\n"
	"Note : the values are expanded as-is in a one time pass.\n"
	"The history will save the expanded line, not with the $VARIABLE\n"
	"A short example: \n"
	"\% env -s ADDR 0xc0123456\n"
	"\% kmem -w $ADDR \\x44\\x45\\x40\\x44\n"
	"\n", stdout );
}

void
do_env( int pargc, char * pargv[] )
{
	char c = 0;
	char * value = NULL;
	int i = 0;
	int size = 0;

	NOARGUMENTS( pargc, 1, set );

	optarg = NULL; /* If I don't do this I can call it only one time ... */
	optind = 0;

	c = getopt( pargc, pargv, "clsa::" ); 
	{
		switch ( c ) {
			case 'c':
				clear_all( env.next );
				break;
			case 's':
				for ( i = 3; i < pargc; i++ ) {
					if ( value != NULL ) {
						size = strlen( value );
					}
					value = realloc( value, strlen(pargv[i])+size+2 );
					if ( value == NULL ) {
						perror( "do_set:realloc()" );
						return;
					}

					if ( size != 0 ) {
						value[size] = ' ';
						strcpy( value+size+1, pargv[i] );
					} else {
						strcpy( value, pargv[i] );
					}
				}
				add_env( pargv[2], value );
				break;
			case 'l':
				list_all();
				break;
			default:
				usage_env();
				return;
		}
	}
	return;
}
