/* 
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <asm/segment.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "include/kernsh.h"
#include "include/interrupt.h"

#define SYSTEM "System gate"
#define INTERRUPT "Interrupt gate"
#define TRAP "Trap gate"
#define NR_INT 256

#define CHECK_NUMBER_INTERRUPT( n ) { \
		if ( n < 0 || n > NR_INT-1 ) { \
				fatal( "interrupt # (%i) not in range\n", n ); \
				return; \
			} \
		}

unsigned char *
get_segment( unsigned short selecteur_interrupt )
{
	if ( selecteur_interrupt == __KERNEL_CS ) {
		return( "KERNEL_CS" );
	}
	if ( selecteur_interrupt == __KERNEL_DS ) {
		return( "KERNEL_DS" );
	}
	if ( selecteur_interrupt == __USER_CS ) {
		return( "USER_CS" );
	}
	if ( selecteur_interrupt == __USER_DS ) {
		return( "USER_DS" );
	}
	return( NULL );
}

unsigned char *
get_idt_segment( unsigned int interrupt )
{
	int n;

	n = read_kmem( &idt, idtr.base+8*interrupt, sizeof(idt),
	file_descriptor_kmem );
	if ( n < 0 ) {
		return( 0 );
	}

	return( get_segment((unsigned short)idt.sel) );
}

unsigned int
get_idt_dpl( void )
{
	if ( idt.flags & 64 ) {
		return( 3 );
	}
	return( 0 );
}

unsigned char *
get_idt_type( unsigned int dpl_interrupt )
{
	unsigned char * type_interrupt = NULL;

	type_interrupt = malloc( sizeof(INTERRUPT)-1 );
	if ( type_interrupt == NULL ) {
		perror( "get_idt_type:malloc()" );
		return( NULL );
	}
	memset( type_interrupt, '\0', sizeof(INTERRUPT)-1 );

	if ( idt.flags & 1 ) {
		if ( dpl_interrupt ) {
			memcpy( type_interrupt, SYSTEM, sizeof(SYSTEM)-1 );
			type_interrupt[ strlen(SYSTEM) ] = '\0';
		}
		else {
			memcpy( type_interrupt, TRAP, sizeof(TRAP)-1 );
			type_interrupt[ strlen(TRAP) ] = '\0';
		}
	}
	else {
		memcpy( type_interrupt, INTERRUPT, sizeof(INTERRUPT)-1 );
		type_interrupt[ strlen(INTERRUPT) ] = '\0';
	}
	return( type_interrupt );
}

void 
modify_idt( unsigned int nbr_interrupt, size_t new_address )
{
	unsigned char * type_interrupt = NULL;
	unsigned char * segment_interrupt = NULL;
	unsigned int dpl_interrupt;

	segment_interrupt = get_idt_segment( nbr_interrupt );
	if ( segment_interrupt == NULL ) {
		fatal( "can't get segment : %d..\n\n", nbr_interrupt );
		return;
	}

	idt.off1 = (unsigned short)( new_address );
	idt.off2 = (unsigned short)( new_address >> 16 );
	idt.flags = 239;

	write_kmem( &idt, idtr.base+8*nbr_interrupt, sizeof(idt), 
	file_descriptor_kmem );

	dpl_interrupt = get_idt_dpl();

	type_interrupt = get_idt_type( dpl_interrupt );
	if ( type_interrupt == NULL ) {
		fatal( "can't get type of interrupt..\n\n" );
		return;
	}

	fputs( "Idt modified..\n", stdout );
	fprintf( stdout, "int     : %d\n", nbr_interrupt ); 
	fprintf( stdout, "address : %#x\n", new_address );
	fprintf( stdout, "segment : %s\n", segment_interrupt );
	fprintf( stdout, "dpl     : %d\n", dpl_interrupt );
	fprintf( stdout, "type    : %s\n\n", type_interrupt );
	return;
}

void
list_idt( void )
{
	unsigned char * type_interrupt = NULL;
	unsigned char address[ 2*sizeof(size_t)+2 ];
	unsigned char * segment_interrupt = NULL;
	unsigned char * symbol_interrupt = NULL;
	char c;
	size_t p_interrupt;
	unsigned int dpl_interrupt;
	int i;
	int j;
	
	fprintf( stdout, "Number of IDT entries: %d\n", NR_INT );
	fputs( "-------------------------------------------------\n\n", 
	stdout );

	for ( i = 0, j = 0; i < NR_INT; i++ ) {
		segment_interrupt = get_idt_segment( i );

		p_interrupt = ( idt.off2 << 16 ) + idt.off1;
		if ( p_interrupt != 0 ) {
			dpl_interrupt = get_idt_dpl();

			type_interrupt = get_idt_type( dpl_interrupt );
			if ( type_interrupt == NULL ) {
				fatal( "can't get type of interrupt..\n\n" );
				return;
			}

			snprintf( address, sizeof(address), "%#x", p_interrupt );
			symbol_interrupt = get_symbol_by_addr( address );

			fprintf( stdout, "IDT Entry (%s) #%d\n", 
			type_interrupt, i );
			fprintf( stdout, "\tflags: %#x, dpl: %d, offset: %#x\n\t"
			"symbol: %s, segment: %s\n\n", idt.flags, dpl_interrupt, 
			p_interrupt, symbol_interrupt, segment_interrupt );
		}

		if ( !script_on_command_line ) {
			j++;
			if ( !((j+1)%10) ) {
				c = enter();
				if ( c == '\r' || c == '\n' ) {
					continue;
				}
				if ( c == 'q' ) {
					getchar();
					return;
				}
			}
		}
	}
	fputs( "\n", stdout );
	return;
}

void
signatures_idt( char * signatures_file, int size_sign )
{
#define CREATE_IDT "checksums MD5 of IDT\n"
	create_file_signatures( CREATE_IDT, signatures_file, size_sign, 
	SIG_FOR_IDT );
	return;
}

/* cf: doc/idt.txt */
void
check_idt( char * signatures_file, int size_sign )
{
	int i;
	
	i = file_is_present( signatures_file );
	if ( i < 0 ) {
		return;
	}

#define CHECK_IDT "\nn\tsignature\t\tactual\n" \
	"---------------------------------------------\n"
	i = check_file_signatures( CHECK_IDT, signatures_file, size_sign, 
	SIG_FOR_IDT );
	if ( i == EXIT_SUCCESS ) {
		fputs( "IDT is valid..\n\n", stdout );
	}
	return;
}

/* Cf. include/interrupt.h */
void
call_interrupt( int interrupt )
{
	if ( interrupt == 0 ) { INTERRUPT_0; } /* divide_error */
	if ( interrupt == 1 ) { INTERRUPT_1; } /* debug */
	if ( interrupt == 2 ) { INTERRUPT_2; } /* nmi */
	if ( interrupt == 3 ) { INTERRUPT_3; } /* int 3 */
	if ( interrupt == 4 ) { INTERRUPT_4; } /* overflow */
	if ( interrupt == 5 ) { INTERRUPT_5; } /* bounds */
	if ( interrupt == 6 ) { INTERRUPT_6; } /* invalid_op */
	if ( interrupt == 7 ) { INTERRUPT_7; } /* device_not_available */
	if ( interrupt == 8 ) { INTERRUPT_8; } /* double_fault */   
	if ( interrupt == 9 ) { INTERRUPT_9; } /* coprocessor_segment_overrun */
	if ( interrupt == 10 ) { INTERRUPT_10; } /* invalid_TSS */
	if ( interrupt == 11 ) { INTERRUPT_11; } /* segment not present */
	if ( interrupt == 12 ) { INTERRUPT_12; } /* stack_segment */
	if ( interrupt == 13 ) { INTERRUPT_13; } /* general_protection */
	if ( interrupt == 14 ) { INTERRUPT_14; } /* page_fault */
	if ( interrupt == 15 ) { INTERRUPT_15; } /* spurious_interrupt_bug */
	if ( interrupt == 16 ) { INTERRUPT_16; } /* coprocessor_error */
	if ( interrupt == 17 ) { INTERRUPT_17; } /* alignnment_check */
	if ( interrupt == 18 ) { INTERRUPT_18; } /* machine_check */
	if ( interrupt == 19 ) { INTERRUPT_19; } /* simd_coprocessor_error */
	if ( interrupt == 250 ) { INTERRUPT_250; } /* ignore_int */
}

void
usage_idt( void )
{
	fputs( 
	"idt [ -lmCca ] <arguments>\n\n"
	"This command makes it possible to handle the interrupts description\n"
	"table (or IDT): reading, writing..\n\n"
	"__options\n"
	"  -l : display IDT ;\n"
	"  -m [int] [address] : modify the interrupt gate of one of the\n"
	"entries of the IDT ;\n" 
	"  -a [file] [size of signatures] : create a file of checksums MD5\n" 
	"of the entries of the IDT ;\n" 
	"  -C [file] [size of signatures] : check IDT integrity ;\n"
	"  -c [int] : call a interrupt among the following ones:\n\n"
	"\t0 - divide error\n"
	"\t1 - debug\n"
	"\t2 - nmi\n"
	"\t3 - int3\n"
	"\t4 - overflow\n"
	"\t5 - bounds\n"
	"\t6 - invalid_op\n"
	"\t7 - device_not_available\n"
	"\t8 - double_fault\n"
	"\t9 - coprocessor_segment_overrun\n"
	"\t10 - invalid_TSS\n"
	"\t11 - segment_not_present\n"
	"\t12 - stack segment\n"
	"\t13 - general_protection\n"
	"\t14 - page_fault\n"
	"\t15 - spurious_interrupt_code\n"
	"\t16 - coprocessor_error\n"
	"\t17 - alignment_check\n"
	"\t18 - machine_check\n"
	"\t19 - simd_coprocessor_error\n"
	"\t250 - ignore_int\n\n", stdout );
	return;
}

#define MODIFY_IDT 0
#define CREATE_SIGNATURE_IDT 1
#define CHECK_SIGNATURE_IDT 2
#define CALL_INTERRUPT 3

void
do_idt( int pargc, char * pargv[] )
{
	char c;
	size_t l;
	unsigned int n;
	int flag;
	int syscall;
	int size_sign = 10; /* 10 is a minimum ! */
	
	NOARGUMENTS( pargc, 1, idt );
	
	optind = 0;
	flag = -1;
	
	while ( (c = getopt(pargc, pargv, "lmcaC")) != EOF ) {
		switch ( c ) {
			case 'l': 
				list_idt();
				break;
			case 'm':
				flag = MODIFY_IDT;
				break;
			case 'a':
				flag = CREATE_SIGNATURE_IDT;
				break;
			case 'C':
				flag = CHECK_SIGNATURE_IDT;
				break;
			case 'c':
				flag = CALL_INTERRUPT;
				break;
			default:
				usage_idt();
				return;
		}
	}

	switch ( flag ) {
		case CREATE_SIGNATURE_IDT:
			if ( pargc < 3 ) {
				usage_idt();
				return;
			}

			if ( pargc == 4 ) {
				size_sign = atoi( pargv[3] );
				if ( size_sign > 64 ) {
					fatal( "please, choose size < 64..\n" );
					return;
				}
			}
			
			signatures_idt( pargv[2], size_sign );	
			break;

		case CHECK_SIGNATURE_IDT:
			if ( pargc < 3 ) {
				usage_idt();
				return;
			}

			if ( pargc == 4 ) {
				size_sign = atoi( pargv[3] );
				if ( size_sign > 64 ) {
					fatal( "please, choose size < 64..\n" );
					return;
				}
			}
			
			check_idt( pargv[2], size_sign );	
			break;

		case MODIFY_IDT:
			if ( pargc != 4 ) {
				usage_idt();
				return;
			}

			n = atoi( pargv[2] );
			CHECK_NUMBER_INTERRUPT( n );

			XSTRTOUL( pargv[3], l );
			modify_idt( n, l );
			break;

		case CALL_INTERRUPT:
			if ( pargc < 3 ) {
				usage_idt();
				return;
			}

			n = atoi( pargv[2] );
			CHECK_NUMBER_INTERRUPT( n );

			if ( n == 128 ) {
				syscall = atoi( pargv[3] );
				CHECK_NUMBER_SYSCALL( syscall );
				call_syscall( syscall );
			} else {
				call_interrupt( n );
			}
			break;

		default:
			return;
	}

	return;
}

	

