#ifndef __INTERRUPT_H
#define __INTERRUPT_H

/* L'utilisation de l'instruction int nous empche d'utiliser des
 * variables (elle n'accepte que des chanes constantes). Ce qui 
 * fait que nous dclarons une macro pour chaque interruption ou
 * presque (c'est probable qu'il y est une meilleure solution). 
 * Nous ne prenons pas en compte ni les interruptions matrielles, 
 * ni l'interruption system_call (l'appel  un syscall est dj 
 * implment). Et pour toutes les interruptions non implmentes 
 * (qui ont pour handler ignore_int()), nous utilisons seulement 
 * l'interruption 250 (inutile de toutes les dclarer).
 *
 * Ceci dit, si vous voulez en dclarer plus, libre  vous de le 
 * faire, cela n'est pas vraiment compliqu. 
 */

/* Interrupt divide_error */
#define INTERRUPT_0 { \
		long __res; \
		__asm__ volatile ( "int $0x0" \
						: "=a" (__res) ); \
		} 

/* Interrupt debug */
#define INTERRUPT_1 { \
		long __res; \
		__asm__ volatile ( "int $0x1" \
						: "=a" (__res) ); \
		} 

/* Interrupt nmi */
#define INTERRUPT_2 { \
		long __res; \
		__asm__ volatile ( "int $0x2" \
						: "=a" (__res) ); \
		} 

/* Interrupt int3 */
#define INTERRUPT_3 { \
		long __res; \
		__asm__ volatile ( "int $0x3" \
						: "=a" (__res) ); \
		} 

/* Interrupt overflow */
#define INTERRUPT_4 { \
		long __res; \
		__asm__ volatile ( "int $0x4" \
						: "=a" (__res) ); \
		} 

/* Interrupt bounds */
#define INTERRUPT_5 { \
		long __res; \
		__asm__ volatile ( "int $0x5" \
						: "=a" (__res) ); \
		} 

/* Interrupt invalid_op */
#define INTERRUPT_6 { \
		long __res; \
		__asm__ volatile ( "int $0x6" \
						: "=a" (__res) ); \
		} 

/* Interrupt device_not_available */
#define INTERRUPT_7 { \
		long __res; \
		__asm__ volatile ( "int $0x7" \
						: "=a" (__res) ); \
		} 

/* Interrupt double_fault */
#define INTERRUPT_8 { \
		long __res; \
		__asm__ volatile ( "int $0x8" \
						: "=a" (__res) ); \
		} 

/* Interrupt coprocessor_segment_overrun */
#define INTERRUPT_9 { \
		long __res; \
		__asm__ volatile ( "int $0x9" \
						: "=a" (__res) ); \
		} 


/* Interrupt invalid_TSS */
#define INTERRUPT_10 { \
		long __res; \
		__asm__ volatile ( "int $0xa" \
						: "=a" (__res) ); \
		} 

/* Interrupt segment_not_present */
#define INTERRUPT_11 { \
		long __res; \
		__asm__ volatile ( "int $0xb" \
						: "=a" (__res) ); \
		} 

/* Interrupt stack_segment */
#define INTERRUPT_12 { \
		long __res; \
		__asm__ volatile ( "int $0xc" \
						: "=a" (__res) ); \
		} 

/* Interrupt general_protection */
#define INTERRUPT_13 { \
		long __res; \
		__asm__ volatile ( "int $0xd" \
						: "=a" (__res) ); \
		} 

/* Interrupt page_fault */
#define INTERRUPT_14 { \
		long __res; \
		__asm__ volatile ( "int $0xe" \
						: "=a" (__res) ); \
		} 

/* Interrupt spurious_interrupt_bug */
#define INTERRUPT_15 { \
		long __res; \
		__asm__ volatile ( "int $0xf" \
						: "=a" (__res) ); \
		} 

/* Interrupt coprocessor_error */
#define INTERRUPT_16 { \
		long __res; \
		__asm__ volatile ( "int $0x10" \
						: "=a" (__res) ); \
		} 

/* Interrupt alignment_check */
#define INTERRUPT_17 { \
		long __res; \
		__asm__ volatile ( "int $0x11" \
						: "=a" (__res) ); \
		} 

/* Interrupt machine_check */
#define INTERRUPT_18 { \
		long __res; \
		__asm__ volatile ( "int $0x12" \
						: "=a" (__res) ); \
		} 

/* Interrupt simd_coprocessor_error */
#define INTERRUPT_19 { \
		long __res; \
		__asm__ volatile ( "int $0x13" \
						: "=a" (__res) ); \
		} 

/* Interrupt ignore_int */
#define INTERRUPT_250 { \
		long __res; \
		__asm__ volatile ( "int $0xfa" \
						: "=a" (__res) ); \
		} 

#endif /* __INTERRUPT_H */
