/* 
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "include/kernsh.h"
#include <readline/readline.h>

char *
dupstr( char * s )
{
	char * r = NULL;

	r = xmalloc( strlen(s)+1 );
	strcpy( r, s );
	return( r );
}

/* failure notice and exit */
void 
fatal( const char * format, ... ) 
{
	va_list ap;
	
	va_start( ap, format );
	vfprintf( stderr, format, ap );
	fflush( stderr );
	va_end( ap );
	return;
}

/* thanks kraken for hexdump function :-) */
void
hexdump( unsigned char * data, unsigned int amount, size_t addr )
{
	unsigned int dp;
	unsigned int p;
	const char trans[] =
	"................................ !\"#$%&'()*+,-./0123456789"
	":;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklm"
	"nopqrstuvwxyz{|}~...................................."
	"....................................................."
	"........................................";
	
	for ( dp = 1; dp <= amount; dp++ ) {
        if ( (dp % 16) == 1 ) {
            fprintf( stdout, "%#08x | ", addr+dp-1 );
        }

        fprintf( stdout, "%02x ", data[dp-1] );
        if ( (dp % 8) == 0 && (dp % 16) != 0 ) {
            fputs( " ", stdout );
        }
        if ( (dp % 16) == 0 ) {
            fputs( "| ", stdout );
            p = dp;
            for ( dp -= 16; dp < p; dp++ ) {
                fprintf( stdout, "%c", trans[data[dp]] );
            }
            fputs( "\n", stdout );
        }
    }
	
	if ( (amount % 16) != 0 ) {
		p = dp = 16 - ( amount % 16 );
		for ( dp = p; dp > 0; dp-- ) {
			fputs( "   ", stdout );
			if ( ((dp % 8) == 0) && (p != 8) ) {
				fputs( " ", stdout );
			}
		}
		fputs( "| ", stdout );
		for ( dp = (amount - (16 - p)); dp < amount; dp++ ) {
			fprintf( stdout, "%c", trans[data[dp]] );
		}
	}

	fputs( "\n", stdout );
	return;
}

char
enter( void )
{
    fputs( "---Type <return> to continue, or q <return> to quit---\n", 
		stderr );
    return( getchar() );
}

/* strip whitespace from the start and end of STRING.  Return a pointer
 * into STRING. 
 */
char *
stripwhite( char * string )
{
	register char * s = NULL; 
	register char * t = NULL;

	for ( s = string; whitespace(*s); s++ );
	if ( *s == 0 ) {
		return( s );
	}

	t = s + strlen( s ) - 1;
	while ( t > s && whitespace(*t) ) {
		t--;
	}
	*++t = '\0';
	return( s );
}

char *
string2hex( char * str, int * nu )
{
	char * pointer = NULL;
	char tmp[ 5 ];
	int i;
	int a;
	int len;

	len = strlen( str ) + 1;
	a = len / 4;
	*nu = a;

	pointer = malloc( a+1 );
	if ( pointer == NULL ) {
		perror( "string2hex:malloc()" );
		return( NULL );
	}
	memset( pointer, '\0', a+1 );

	for ( i = a = 0; i < len; i += 4 ) {
		if ( str[i] == '\\' ) {
			strcpy( tmp, "0x" );
			strncat( tmp, &str[i+2], 2 );
			tmp[ sizeof(tmp)-1 ] = 0;
			pointer[a] = strtoul( tmp, NULL, 16 );
			a++;
			if ( a == len/4 ) {
				break;
			}
		}
	}
	
	if ( !a ) {
		fatal( "string %s not valid..\n\n", str );
		return( NULL );
	}
	return( pointer );
}

void *
xmalloc( size_t size )
{
	char * buf = NULL;

	buf = malloc( size );
	if ( buf == NULL ) {
		return( NULL );
	}

	memset( buf, 0, size );
	return( buf );
}

int
file_is_present( char * pathname )
{
	struct stat stat_pathname;
	int i;

	i = stat( pathname, &stat_pathname );
	if ( i < 0 ) {
		fatal( "%s doesn't exist probably!\n", pathname );
		return( -1 );
	}
	
	if ( !stat_pathname.st_size ) {
		fatal( "%s is blank !\n", pathname );
		return( -1 );
	}

	return( 0 );
}
