/* 
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "include/kernsh.h"
#include "include/sysc.h"

int
set_system_call( void )
{
	int n = 0;

	/* get idt */
#ifndef DEBUG
	asm( "sidt %0" : "=m" (idtr) );
	/* get system_call address */
	n = read_kmem( &idt, idtr.base + 8 * 0x80, sizeof(idt),
		file_descriptor_kmem );
	if ( n < 0 ) {
		return( n );
	}
	p_system_call = (idt.off2 << 16) | idt.off1;
#else
	n = 0;
	p_system_call = get_symbol_by_name( "system_call" );
#endif

	if ( p_system_call != get_symbol_by_name("system_call") ) {
		fprintf( stdout, "\n[#] The IDT is hijacked ! Fake system_call"
			" at %#x\n", p_system_call );
		fputs( "[#] Your System.map file is perhaps bad\n", stdout );
		fputs( "[#] Create it with the script scripts/create_system_map.sh\n\n", stdout );
		return( 1 );
	}
	return( 0 );
}

int
set_syscall_table( void )
{
	int n = 0;
	size_t offset = 0;
	char * here = NULL;
	char buf[ BUFSIZ ];

	/* get sys_call_table address */
	/* the following method should work for 2.2/2.4/2.6 kernel */
	n = read_kmem( buf, p_system_call, sizeof(buf), file_descriptor_kmem );
	if ( n < 0 ) {
		return( -1 );
	}

	here = (char *)memmem( buf, sizeof(buf), "\xff\x14\x85", 3 );
	if ( here == NULL ) {
		perror( "set_syscall_table:memmem()");
		return( -1 );
	}
	offset = (int)( here - p_system_call ) + 3;

	n = read_kmem( &p_sys_call_table, p_system_call+offset, 4, 
		file_descriptor_kmem );
	if ( n < 0 ) {
		return( n );
	}

	if ( p_sys_call_table != get_symbol_by_name("sys_call_table") ) {
		fatal( "[#] system_call function hijacked ! "
			"new sys_call_table at %#x\n", p_sys_call_table );
		fatal( "[#] trying to work anyway..\n\n" );
		return( 1 );
	}
	return( 0 );
}

void
get_list_addr_syscall( void )
{
	char buf[ BUFSIZ ];
	char buf2[ BUFSIZ ];
	int i;

	for ( i = 0; i < NR_syscalls; i+=2 ) {
		snprintf( buf, sizeof(buf), "[%3i] %20s: 0x%08lx", i, list_syscall[i],
			(unsigned long)sys_call_table[i] );
		snprintf( buf2, sizeof(buf2), "[%3i] %20s: 0x%08lx", i+1, list_syscall[i+1], 
			(unsigned long)sys_call_table[i+1] );
		if ( strlen(buf)+strlen (buf2)+2 <= 80 ) {
			fprintf( stdout, "%s  %s\n", buf, buf2 );
		} else {
			fprintf( stdout, "%s\n%s\n", buf, buf2 );
		}
	}
	return;
}

void
call_syscall( int number )
{
	long __res;

	__asm__ volatile( "int $0x80"
		: "=a" (__res)
		: "0" (number) );
	return;
}

int
refresh_syscall_table( void )
{
	int n;
	
	n = read_kmem( &sys_call_table, p_sys_call_table,
		sizeof(sys_call_table), file_descriptor_kmem );
	return( n );
}

void
hijack_syscall( int number, size_t kaddr )
{
	int n;

	n = write_kmem( &kaddr, p_sys_call_table+(sizeof(void *)*number),
		sizeof(kaddr), file_descriptor_kmem );
	if ( n < 0 ) {
		fatal( "writing in %s failed !\n", KMEM );
	}

	/* sys_call_table is modified, so we initialize again 
	 * sys_call_table 
	 */
	n = refresh_syscall_table();
	if ( n < 0 ) {
		return;
	}
	return;
}

/* That is not enough to check the first 7 bits representing a jmp. 
 * I thus prefer to make a file of signatures of the syscalls on a 
 * healthy system and to make a checking starting from this check.
 */
void
signatures_syscall( char * signatures_file, int size_sign )
{
#define CREATE_SYSCALL "signature of syscalls\n"
	create_file_signatures( CREATE_SYSCALL, signatures_file, size_sign,
	SIG_FOR_SYSCALL );
	return;
}

void
check_syscall( char * signatures_file, int size_sign )
{
	int i;
	
	i = file_is_present( signatures_file );
	if ( i < 0 ) {
		return;
	}

#define CHECK_SYSCALL "\nn\tsignature\t\tactual\n" \
	"----------------------------------------------------------\n"
	i = check_file_signatures( CHECK_SYSCALL, signatures_file, size_sign,
		SIG_FOR_SYSCALL );
	if ( i == EXIT_SUCCESS ) {
		fputs( "syscall_table valid..\n\n", stdout );
	}
	return;
}

static size_t start_address;
static size_t end_address;

void
get_definition_syscall( int nr )
{
	fputs( "         name : ", stdout );
	if ( !strlen(list_syscall[nr]) ) {
		fputs( "(null)\n", stdout );
	}
	else {
		fprintf( stdout, "%s\n", list_syscall[nr] );
	}
	
	fprintf( stdout, "   syscall n : %d\n", nr );
	start_address = (unsigned int)sys_call_table[nr];
	fprintf( stdout, "start address : %#x\n", start_address );

	end_address = get_next_symbol_by_addr( start_address );
	if ( end_address != 0 ) {
		fprintf( stdout, "  end address : %#x\n", end_address );
		fprintf( stdout, "       length : %d\n", 
			end_address-start_address );
		fputs( "\n", stdout );
	}
	return;
}

void
get_definition_syscall_by_number( int nr )
{
	get_definition_syscall( nr );
}

void
get_definition_syscall_by_addr( char * address )
{
	size_t p_syscall;
	int i;
	int n;
	int nr = 0;

	p_syscall = strtoul( address, NULL, 16 );
	if ( p_syscall < PAGE_OFFSET || p_syscall > 0xffffffff ) {
		fatal( "address not valid..\n" );
		return;
	}

	for ( i = 0, n = 0; i < NR_syscalls; i++ ) {
		if ( p_syscall == (unsigned int)sys_call_table[i] ) {
			nr = i;
			n++;
			if ( n > 1 ) {
				fprintf( stdout, "address of syscall "
				"sys_ni_syscall..\n\n" );
				return;
			}
		}
	}
    
	if ( nr ) get_definition_syscall( nr );
	else fatal( "syscall at %#x() don't exist..\n", p_syscall );

	return;
}

void
get_definition_syscall_by_name( char * name_syscall )
{
	int i;
	int nr = 0;

	for ( i = 0; i < NR_syscalls; i++ ) {
		if ( !strcmp(list_syscall[i], name_syscall) ) {
			nr = i;
		}
	}

	if ( nr ) get_definition_syscall( nr );
	else fatal( "syscall %s don't exist..\n\n", name_syscall );

	return;
}

void
get_strange_syscall( void )
{
	int i = 0;
	int positive = 0 ;

	if ( refresh_syscall_table () < 0 ) {
		perror( "get_stange_syscall:refresh_syscall_table()" );
		return;
	}

	for ( i = 0; i < NR_syscalls; i++ ) {
		if ( (unsigned int) sys_call_table[i] < PAGE_OFFSET + 0x00100000
			|| (unsigned int) sys_call_table[i] > PAGE_OFFSET + 0x00200000 ) {
			fprintf( stdout, "WARNING : syscall n %d not in the good place : %#08lx\n",
				i, sys_call_table[i] );
			positive++;
			continue;
		}
	}

	if ( !positive ) {
		fprintf( stdout,"All tests passed !\n");
	}
	return;
}

void
usage_syscall( void )
{
	fputs( 
	"syscall [ -dangshlrct ] <arguments>\n\n"
	"This command makes it possible to play with the syscalls:\n"
	"list syscalls and their addresses, syscall hijack..\n\n"
	"__options\n"
	"  -d [syscall number] <opcode|disass> : display characteristics\n" 
	"of a syscall according to a number ;\n" 
	"  -a [address] <opcode|disass> : display characteristics of a\n"
	"syscall according to an address ;\n"
	"  -n <syscall name> <opcode|disass> : display characteristics of\n" 
	"a syscall according to a name ;\n"
	"  -g [filename] <size_of_signature> : create a file of signatures ;\n"
	"  -s [filename] <size_of_signature> : check syscall_table integrity ;\n"
	"  -h [syscall number] <syscall address> : hijack a syscall ;\n"
	"  -l : list syscalls and theirs addresses ;\n"
	"  -r [ 0 | 1 | 2 ] : manually refresh :\n"
	"\t0 : syscall_table contents ;\n"
	"\t1 : syscall_table address ;\n"
	"\t2 : system_call address.\n"
	"  -t : check for strange syscall ;\n"
	"  -c [syscall number] : call a syscall according to a number ;\n\n",
	stdout );
	return;
}

#define HIJACK_SYSCALL 0
#define CREATE_SIGNATURE_SYSCALL 1
#define CHECK_SIGNATURE_SYSCALL 2

void
do_syscall( int pargc, char * pargv[] )
{
	char c;
	size_t l;
	int size_sign = 10; /* 10 by default is a minimum ! */
	int flag;
	int i;

	NOARGUMENTS( pargc, 1, syscall );

	start_address = end_address = 0;
	optind = 0;
	flag = -1;

	while ( (c = getopt(pargc, pargv, "htlr:s:g:d:c:a:n:")) != EOF ) {
		switch ( c ) {
			case 't':
				get_strange_syscall();
				break;
			case 'r':
				switch( optarg[0] ) {
				case '0':
					refresh_syscall_table();
					break;
				case '1':
					set_syscall_table();
					break;
				case '2':
					set_system_call();
					break;
				default :
					set_syscall_table();
					set_system_call();
					refresh_syscall_table();
					break;
				}
				break;
			case 'l':
				get_list_addr_syscall();
				break;
			case 'c':
				i = atoi( optarg );
				CHECK_NUMBER_SYSCALL( i );
				call_syscall( i );
				break;
			case 's':
				flag = CHECK_SIGNATURE_SYSCALL;
				break;
			case 'g':
				flag = CREATE_SIGNATURE_SYSCALL;
				break;
			case 'h':
				flag = HIJACK_SYSCALL;
				break;
			case 'd':
				i = atoi( optarg );
				CHECK_NUMBER_SYSCALL( i );
				get_definition_syscall_by_number( i );
				break;
			case 'a':
				get_definition_syscall_by_addr( optarg );
				break;
			case 'n':
				get_definition_syscall_by_name( optarg );
				break;
			default:
				usage_syscall();
				return;
		}
	}

    if ( pargv[3] && start_address != 0 && end_address != 0 ) {
        if ( !strcmp(pargv[3], "opcode") ) {
        fprintf( stdout, "[!] opcode from address %#x..\n",
            start_address );
        kdump( start_address, end_address-start_address,
            file_descriptor_kmem );
        }
        if ( !strcmp(pargv[3], "disass") ) {
        fprintf( stdout, "[!] disassembling from address %#x..\n\n",
            start_address );
        /*i = disasm( start_address, end_address-start_address );
        if ( i < 0 ) {
        fatal( "can't disass this symbol..\n" );
        }
		*/
		kdisass( start_address, end_address-start_address, 
			file_descriptor_kmem );
		fputs( "\n", stdout );
        }
		start_address = end_address = 0; /* que diable ! */
		return;
    }

	switch ( flag ) {
		case HIJACK_SYSCALL:
			if ( pargc != 4 ) { 
				usage_syscall();
				return;
			}
			i = atoi( pargv[2] );
			CHECK_NUMBER_SYSCALL( i );
			XSTRTOUL( pargv[3], l );

			fprintf( stdout, "hijacked syscall : %d\n", i );
			fprintf( stdout, "    with address : %#x\n\n", l );
			hijack_syscall( i, l );
			break;

		case CHECK_SIGNATURE_SYSCALL:
			if ( pargc < 3 ) {
				usage_idt();
				return;
			}

			if ( pargc == 4 ) {
				size_sign = atoi( pargv[3] );
				if ( size_sign > 64 ) {
					fatal( "please, choose size < 64..\n" );
					return;
				}
			}
			check_syscall( pargv[2], size_sign );
			break;

		case CREATE_SIGNATURE_SYSCALL:
			if ( pargc < 3 ) {
				usage_idt();
				return;
			}

			if ( pargc == 4 ) {
				size_sign = atoi( pargv[3] );
				if ( size_sign > 64 ) {
					fatal( "please, choose size < 64..\n" );
					return;
				}
			}
			signatures_syscall( pargv[2], size_sign );
			break;

		default:
			return;
	}
	return;
}
