/* 
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <unistd.h>
#include "../libdisasm/libdis.h"
#include "include/kernsh.h"

/* XXX Should be move to a common header with modules/lkm-virtm.o */
#define __NR_getvirtaddr 251
#define __NR_rwvirtaddr 250

int
parse_getvirtaddr( struct mem_addr * mem, int pid )
{
	int __ret;

	SYSCALL2( __NR_getvirtaddr, mem, pid );
	return __ret;
}

int
parse_rwvirtaddr( int pid, char * buf, int size, size_t addr, int what )
{
	int __ret;
	
	SYSCALL5( __NR_rwvirtaddr, pid, buf, size, addr, what );
	return __ret;
}

void
virtm_error_message( int i )
{
	if ( i == 1 ) fatal( "is the pid correct ?\n" );
	if ( i == -1 ) fatal( "out of memory\n" );
	if ( i == -2 ) fatal( "invalid or reserved page\n" );
}

void
get_virtual_memory( pid_t pid )
{
	struct mem_addr addresses;
    int i;
	
    i = parse_getvirtaddr( &addresses, pid );
    if ( i ) {
		virtm_error_message( i );
		return;
    }

    fprintf( stdout, "[#] virtual memory for pid %d:\n\n", pid );
    fprintf( stdout, "vm_start(1)       : \t%#x\n", (size_t)addresses.vm_start );
    fprintf( stdout, "vm_end(1)         : \t%#x\n", (size_t)addresses.vm_end );
    fprintf( stdout, "vm_start_cache(1) : \t%#x\n", (size_t)addresses.vm_start_cache );
    fprintf( stdout, "vm_end_cache(1)   : \t%#x\n", (size_t)addresses.vm_end_cache );
    fprintf( stdout, "map_count         : \t%#x\n", (size_t)addresses.map_count );
    fprintf( stdout, "code_start        : \t%#x\n", (size_t)addresses.startcode );
    fprintf( stdout, "code_end          : \t%#x\n", (size_t)addresses.endcode );
    fprintf( stdout, "data_start        : \t%#x\n", (size_t)addresses.startdata );
    fprintf( stdout, "data_end          : \t%#x\n", (size_t)addresses.enddata );
    fprintf( stdout, "heap_start        : \t%#x\n", (size_t)addresses.startbrk );
    fprintf( stdout, "stack_start       : \t%#x\n", (size_t)addresses.startstack );
    fprintf( stdout, "arg_start         : \t%#x\n", (size_t)addresses.arg_start );
    fprintf( stdout, "arg_end           : \t%#x\n", (size_t)addresses.arg_end );
    fprintf( stdout, "env_start         : \t%#x\n", (size_t)addresses.env_start );
    fprintf( stdout, "env_end           : \t%#x\n", (size_t)addresses.env_end );
    fprintf( stdout, "rss               : \t%#x\n", (size_t)addresses.rss );
    fprintf( stdout, "total_vm          : \t%#x\n", (size_t)addresses.total_vm );
    fprintf( stdout, "locked_vm         : \t%#x\n", (size_t)addresses.locked_vm );
    fprintf( stdout, "swap_address      : \t%#x\n", (size_t)addresses.swap_address );
}

#define READ 0
#define WRITE 1

void
read_virtual_memory( pid_t pid, size_t addr, int ssize )
{
    char * buffer = NULL;
    int i;

    buffer = malloc( ssize );
    if ( buffer == NULL ) {
        perror( "read_virtual_memory:malloc()" );
        return;
    }

    i = parse_rwvirtaddr( pid, buffer, ssize, addr, READ );
    if ( i ) {
		virtm_error_message( i );
        return;
    }

    fprintf( stdout, "\n[ %#x -> %#x | pid : %d ]\n", addr,
        addr+ssize, pid );
    hexdump( buffer, ssize, addr );
    free( buffer );
    fputs( "\n", stdout );
    return;
}

void
write_virtual_memory( pid_t pid, size_t addr, char * buffer )
{
	char * str2hex = NULL;
	char c;
	int i;

	str2hex = string2hex( buffer, &i );
	if ( str2hex == NULL ) {
		return;
	}

	fputs( "\n[ before ]", stdout );
	read_virtual_memory( pid, addr, i );

	fputs( "[ after ]\n", stdout );
	hexdump( str2hex, i, addr );

	fputs( "\nreally writing [y/n] ? ", stdout );
	c = getchar();
	getchar();
	if ( c == 'y' || c == 'Y' ) {
    	i = parse_rwvirtaddr( pid, str2hex, strlen(str2hex), 
			addr, WRITE );
    	if ( i ) {
			virtm_error_message( i );
			return;
    	}
		fprintf( stdout, "strings %s written\n", buffer );
	} else {
		fatal( "can't write on a part of virtual memory\n" );
	}

	fputs( "\n", stdout );
	return;
}

void
search_virtual_memory( int pid, size_t lower, size_t upper, 
	char * string )
{
	char * base = NULL;
	char * found = NULL;
	size_t size;
	int i;

	size = upper - lower;

	base = malloc( size );
	if ( base == NULL ) {
		perror( "search_virtual_memory:malloc()" );
	}

	i = parse_rwvirtaddr( pid, base, size, lower, READ );
	if ( i ) {
		virtm_error_message( i );
		return;
	}

	found = (char *)memmem( base, size, string, strlen(string) );
	if ( found == NULL ) {
		fatal( "no pattern found\n\n" );
		return;
	}
	
	fprintf( stdout, "pattern found at offset %d - virtual address %#x\n\n",
		(size_t)found-(size_t)base, lower+((size_t)found-(size_t)base) );
	return;
}	
	
void
disassemble_virtual_memory( pid_t pid, size_t startaddr, int ssize )
{
    struct mem_addr memory;
	unsigned char * image = NULL;
    int i;

    i = parse_getvirtaddr( &memory, pid );
    if ( i ) {
        fatal( "can't get virtual memory for pid %d\n", pid );
        return;
    }

	image = malloc( ssize );
	if ( image == NULL ) {
		perror( "disassemble_virtual_memory:malloc()" );
		return;
	}
	
	i = parse_rwvirtaddr( pid, image, ssize, startaddr, READ );
	if ( i ) {
		virtm_error_message( i );
		return;
	}

	fprintf( stdout, "[!] disassembling pid %d from address %#x..\n", 
		pid, startaddr );
	fprintf( stdout, "[!] end address: %#x\n", startaddr+ssize );
	fprintf( stdout, "[!] len: %d\n\n", ssize );
	display_code_asm( startaddr, image, ssize );
	return;
}	


void
usage_virtm( void )
{
	fputs( "virtm [ -sgrwd ] <arguments>\n\n"
	"This command makes it possible to read, search, dump and write the virtual\n"
	"address space of a process running.\n\n"
	"__options\n"
	"  -g [pid] : get virtual addresses on the queried process ;\n"
	"  -r [pid] [start address] [size] : read a part of virtual memory ;\n"
	"  -w [pid] [address] [string] : write on a part of virtual memory ;\n"
	"  -s [pid] [start address] [end address] [string] : search a particular\n" 
	"virtual memory area ;\n"
	"  -d [pid] [start address] [size] : disassemble the queried process ;\n\n",
	stdout );
	return;
}

enum {
	READ_VIRTUAL_MEMORY,
	WRITE_VIRTUAL_MEMORY,
	SEARCH_VIRTUAL_MEMORY,
	DISASS_VIRTUAL_MEMORY,
};

void
do_virtm( int pargc, char * pargv[] )
{
	struct module_info info;
	char c;
	size_t l;
	size_t ll;
	size_t ret;
	int flag;
	int i;

#define LKMVIRTM "lkm-virtm"
	i = query_module( LKMVIRTM, QM_INFO, &info, sizeof(info), &ret );
	if ( i ) {
		fatal( "module modules/lkm-virtm.o or an another module not loaded\n" );
		return;
	}
	
	flag = -1;
	optind = 0;
	
	NOARGUMENTS( pargc, 1, virtm );

	while ( (c = getopt(pargc, pargv, "wrsg:d:")) != EOF ) {
		switch ( c ) {
			case 'g': 
				get_virtual_memory( atoi(optarg) );
				break;
			case 'r':
				flag = READ_VIRTUAL_MEMORY;
				break;
			case 'w':
				flag = WRITE_VIRTUAL_MEMORY;
				break;
			case 's':
				flag = SEARCH_VIRTUAL_MEMORY;
				break;
			case 'd':
				flag = DISASS_VIRTUAL_MEMORY;
				break;
			default:
				usage_virtm();
				return;
		}
	}

	switch ( flag ) {
	case READ_VIRTUAL_MEMORY:
		if ( pargc != 5 ) {
			usage_virtm();
			return;
		}

		XSTRTOUL( pargv[3], l );

		read_virtual_memory( atoi(pargv[2]), l, atoi(pargv[4]) );
		break;

	case WRITE_VIRTUAL_MEMORY:
		if ( pargc != 5 ) {
			usage_virtm();
			return;
		}

		XSTRTOUL( pargv[3], l );

		write_virtual_memory( atoi(pargv[2]), l, pargv[4] );
		break;

	case SEARCH_VIRTUAL_MEMORY:
		if ( pargc != 6 ) {
			usage_virtm();
			return;
		}

		XSTRTOUL( pargv[3], l );
		XSTRTOUL( pargv[4], ll );
        if ( ll - l < 0 ) {
            fatal( "size of memory area you want to scan "
                "is negative !\n" );
            return;
        }

		search_virtual_memory( atoi(pargv[2]), l, ll, pargv[5] );
		break;

	case DISASS_VIRTUAL_MEMORY:
		if ( pargc != 5 ) {
			usage_virtm();
			return;
		}

		XSTRTOUL( pargv[3], l );
		disassemble_virtual_memory( atoi(pargv[2]), l, atoi(pargv[4]) );
		break;

	default:
		return;
	}

	return;
}
